@@ -10,12 +10,16 @@
 #ifndef XEUS_JSON_HPP
 #define XEUS_JSON_HPP
 
-#include "nlohmann/json.hpp"
-
-namespace xeus
-{
-    // This is deprecated and will be removed
-    using xjson = nlohmann::json;
-}
+// Remove annoying false positive warning on GCC: json.hpp:1394:23: warning: potential null pointer dereference
+// See https://github.com/nlohmann/json/issues/3525
+#ifdef __GNUC__
+#   pragma GCC diagnostic push
+#   pragma GCC diagnostic ignored "-Wnull-dereference"
+#   include "nlohmann/json.hpp"
+#   pragma GCC diagnostic pop
+#else
+#   include "nlohmann/json.hpp"
+#endif
 
 #endif
+
