@@ -14,7 +14,10 @@
 
 namespace xeus
 {
-    std::string blocking_input_request(const std::string& prompt, bool password)
+    std::string blocking_input_request(
+        const std::string& prompt, 
+        bool password
+    )
     {
         auto& interpreter = xeus::get_interpreter();
 
