@@ -18,6 +18,7 @@
 
 #include "xkernel_core.hpp"
 #include "xeus/xhistory_manager.hpp"
+#include "xeus/xrequest_context.hpp"
 
 using namespace std::placeholders;
 
@@ -40,23 +41,21 @@ namespace xeus
         , p_interpreter(interpreter)
         , p_history_manager(history_manager)
         , p_debugger(debugger)
-        , m_parent_id({guid_list(0), guid_list(0)})
-        , m_parent_header({nl::json::object(), nl::json::object()})
     {
-        // Request handlers
-        m_handler["execute_request"] = &xkernel_core::execute_request;
-        m_handler["complete_request"] = &xkernel_core::complete_request;
-        m_handler["inspect_request"] = &xkernel_core::inspect_request;
-        m_handler["history_request"] = &xkernel_core::history_request;
-        m_handler["is_complete_request"] = &xkernel_core::is_complete_request;
-        m_handler["comm_info_request"] = &xkernel_core::comm_info_request;
-        m_handler["comm_open"] = &xkernel_core::comm_open;
-        m_handler["comm_close"] = &xkernel_core::comm_close;
-        m_handler["comm_msg"] = &xkernel_core::comm_msg;
-        m_handler["kernel_info_request"] = &xkernel_core::kernel_info_request;
-        m_handler["shutdown_request"] = &xkernel_core::shutdown_request;
-        m_handler["interrupt_request"] = &xkernel_core::interrupt_request;
-        m_handler["debug_request"] = &xkernel_core::debug_request;
+        // Request handlers (all but execute_request are blocking)
+        m_handler["execute_request"] = handler_type{&xkernel_core::execute_request, /*blocking*/ false};
+        m_handler["complete_request"] = handler_type{&xkernel_core::complete_request, true};
+        m_handler["inspect_request"] = handler_type{&xkernel_core::inspect_request, true};
+        m_handler["history_request"] = handler_type{&xkernel_core::history_request, true};
+        m_handler["is_complete_request"] = handler_type{&xkernel_core::is_complete_request, true};
+        m_handler["comm_info_request"] = handler_type{&xkernel_core::comm_info_request, true};
+        m_handler["comm_open"] = handler_type{&xkernel_core::comm_open, true};
+        m_handler["comm_close"] = handler_type{&xkernel_core::comm_close, true};
+        m_handler["comm_msg"] = handler_type{&xkernel_core::comm_msg, true};
+        m_handler["kernel_info_request"] = handler_type{&xkernel_core::kernel_info_request, true};
+        m_handler["shutdown_request"] = handler_type{&xkernel_core::shutdown_request, true};
+        m_handler["interrupt_request"] = handler_type{&xkernel_core::interrupt_request, true};
+        m_handler["debug_request"] = handler_type{&xkernel_core::debug_request, true};
 
         // Server bindings
         p_server->register_shell_listener(std::bind(&xkernel_core::dispatch_shell, this, _1));
@@ -65,19 +64,27 @@ namespace xeus
         p_server->register_internal_listener(std::bind(&xkernel_core::dispatch_internal, this, _1));
 
         // Interpreter bindings
-        p_interpreter->register_publisher([this](const std::string& msg_type,
+        p_interpreter->register_publisher([this](xrequest_context request_context,
+                                                 const std::string& msg_type,
                                                  nl::json metadata,
                                                  nl::json content,
                                                  buffer_sequence buffers)
         {
-            this->publish_message(msg_type, std::move(metadata), std::move(content), std::move(buffers),
+            this->publish_message(msg_type, request_context.header(), std::move(metadata), std::move(content), std::move(buffers),
                                   channel::SHELL);
         });
-        p_interpreter->register_stdin_sender(std::bind(&xkernel_core::send_stdin, this, _1, _2, _3));
-        p_interpreter->register_comm_manager(&m_comm_manager);
-        p_interpreter->register_parent_header([this]() -> const nl::json& {
-            return this->parent_header(channel::SHELL);
+
+        p_interpreter->register_stdin_sender([this](xrequest_context request_context,
+                                                   const std::string& msg_type,
+                                                   nl::json metadata,
+                                                   nl::json content)
+        {
+            this->send_stdin(msg_type, request_context.id(), request_context.header(), std::move(metadata), std::move(content));
         });
+
+
+        p_interpreter->register_comm_manager(&m_comm_manager);
+
     }
 
     xkernel_core::~xkernel_core()
@@ -133,6 +140,7 @@ namespace xeus
     }
 
     void xkernel_core::publish_message(const std::string& msg_type,
+                                       nl::json parent_header,
                                        nl::json metadata,
                                        nl::json content,
                                        buffer_sequence buffers,
@@ -140,7 +148,7 @@ namespace xeus
     {
         xpub_message msg(get_topic(msg_type),
                          make_header(msg_type, m_user_name, m_session_id),
-                         get_parent_header(c),
+                         std::move(parent_header),
                          std::move(metadata),
                          std::move(content),
                          std::move(buffers));
@@ -149,12 +157,14 @@ namespace xeus
     }
 
     void xkernel_core::send_stdin(const std::string& msg_type,
+                                  const guid_list& id_list,
+                                  nl::json parent_header,
                                   nl::json metadata,
                                   nl::json content)
     {
-        xmessage msg(get_parent_id(channel::SHELL),
+        xmessage msg(id_list,
                      make_header(msg_type, m_user_name, m_session_id),
-                     get_parent_header(channel::SHELL),
+                     std::move(parent_header),
                      std::move(metadata),
                      std::move(content),
                      buffer_sequence());
@@ -177,21 +187,22 @@ namespace xeus
         return m_comm_manager;
     }
 
-    const nl::json& xkernel_core::parent_header(channel c) const noexcept
+    const nl::json& xkernel_core::parent_header() const noexcept
     {
-        return m_parent_header[std::size_t(c)];
+        return p_interpreter->parent_header();
     }
 
     void xkernel_core::dispatch(xmessage msg, channel c)
     {
         p_logger->log_received_message(msg, c == channel::SHELL ? xlogger::shell : xlogger::control);
-        const nl::json& header = msg.header();
-        set_parent(msg.identities(), header, c);
-        publish_status("busy", c);
+        // Copy because the msg is moved after, and we may need the header
+        // for publishing the status.
+        nl::json header = msg.header();
+        publish_status(header, "busy", c);
 
         std::string msg_type = header.value("msg_type", "");
         handler_type handler = get_handler(msg_type);
-        if (handler == nullptr)
+        if (handler.fptr == nullptr)
         {
             std::cerr << "ERROR: received unknown message" << std::endl;
             std::cerr << "Message type: " << msg_type << std::endl;
@@ -200,7 +211,7 @@ namespace xeus
         {
             try
             {
-                (this->*handler)(std::move(msg), c);
+                (this->*(handler.fptr))(std::move(msg), c);
             }
             catch (std::exception& e)
             {
@@ -208,49 +219,75 @@ namespace xeus
                 std::cerr << "Message type: " << msg_type << std::endl;
             }
         }
-
-        publish_status("idle", c);
+        // async handlers need to set the idle status themselves
+        if(handler.blocking)
+        {
+            publish_status(header, "idle", c);
+        }
     }
 
     auto xkernel_core::get_handler(const std::string& msg_type) -> handler_type
     {
         auto iter = m_handler.find(msg_type);
-        handler_type res = (iter == m_handler.end()) ? nullptr : iter->second;
+        handler_type res = (iter == m_handler.end()) ? handler_type{nullptr} : iter->second;
         return res;
     }
 
-    void xkernel_core::execute_request(xmessage request, channel c)
+    void xkernel_core::execute_request(xmessage request, channel)
     {
+        // xeus assumes execute_request will be executed on SHELL only
         try
         {
             const nl::json& content = request.content();
             std::string code = content.value("code", "");
             bool silent = content.value("silent", false);
             bool store_history = content.value("store_history", true);
-            int execution_count = content.value("execution_count", 1);
             store_history = store_history && !silent;
             nl::json user_expression = content.value("user_expressions", nl::json::object());
             bool allow_stdin = content.value("allow_stdin", true);
             bool stop_on_error = content.value("stop_on_error", false);
 
-            nl::json metadata = get_metadata();
-
-            nl::json reply = p_interpreter->execute_request(
-                code, silent, store_history, std::move(user_expression), allow_stdin);
-
-            std::string status = reply.value("status", "error");
-            send_reply("execute_reply", std::move(metadata), std::move(reply), c);
-
-            if (!silent && store_history)
+            xrequest_context request_context(request.header(), request.identities());
+            execute_request_config config { silent, store_history, allow_stdin };
+            
+            auto reply_callback = [this, request_context, config, stop_on_error, code](nl::json reply)
             {
-                p_history_manager->store_inputs(0, execution_count, code);
-            }
+                int execution_count = 1;
+                execution_count = reply.value("execution_count", 1);
+                std::string status;
+                status = reply.value("status", "error");
+                nl::json metadata = get_metadata();
 
-            if (!silent && status == "error" && stop_on_error)
-            {
-                constexpr long polling_interval = 50;
-                p_server->abort_queue(std::bind(&xkernel_core::abort_request, this, _1), polling_interval);
-            }
+                send_reply(
+                    request_context.id(),
+                    "execute_reply", 
+                    request_context.header(),
+                    std::move(metadata), 
+                    std::move(reply), 
+                    channel::SHELL
+                );
+
+                if (!config.silent && config.store_history)
+                {
+                    p_history_manager->store_inputs(0, execution_count, code);
+                }
+                if (!config.silent && status == "error" && stop_on_error)
+                {
+                    constexpr long polling_interval = 50;
+                    p_server->abort_queue(std::bind(&xkernel_core::abort_request, this, _1), polling_interval);
+                }
+
+                // idle
+                publish_status(request_context.header(), "idle", channel::SHELL);
+            };
+
+            p_interpreter->execute_request(
+                std::move(request_context),
+                std::move(reply_callback),
+                code,
+                config,
+                std::move(user_expression)
+            );
         }
         catch (std::exception& e)
         {
@@ -264,9 +301,9 @@ namespace xeus
         const nl::json& content = request.content();
         std::string code = content.value("code", "");
         int cursor_pos = content.value("cursor_pos", -1);
-
         nl::json reply = p_interpreter->complete_request(code, cursor_pos);
-        send_reply("complete_reply", nl::json::object(), std::move(reply), c);
+        send_reply(request.identities(), "complete_reply", request.header(),
+            nl::json::object(), std::move(reply), c);
     }
 
     void xkernel_core::inspect_request(xmessage request, channel c)
@@ -275,9 +312,8 @@ namespace xeus
         std::string code = content.value("code", "");
         int cursor_pos = content.value("cursor_pos", -1);
         int detail_level = content.value("detail_level", 0);
-
         nl::json reply = p_interpreter->inspect_request(code, cursor_pos, detail_level);
-        send_reply("inspect_reply", nl::json::object(), std::move(reply), c);
+        send_reply(request.identities(), "inspect_reply", request.header(), nl::json::object(), std::move(reply), c);
     }
 
     void xkernel_core::history_request(xmessage request, channel c)
@@ -286,16 +322,15 @@ namespace xeus
 
         nl::json history = p_history_manager->process_request(content);
 
-        send_reply("history_reply", nl::json::object(), std::move(history), c);
+        send_reply(request.identities(), "history_reply",  request.header(), nl::json::object(), std::move(history), c);
     }
 
     void xkernel_core::is_complete_request(xmessage request, channel c)
     {
         const nl::json& content = request.content();
         std::string code = content.value("code", "");
-
         nl::json reply = p_interpreter->is_complete_request(code);
-        send_reply("is_complete_reply", nl::json::object(), std::move(reply), c);
+        send_reply(request.identities(), "is_complete_reply", request.header(), nl::json::object(), std::move(reply), c);
     }
 
     void xkernel_core::comm_info_request(xmessage request, channel c)
@@ -316,14 +351,14 @@ namespace xeus
         nl::json reply;
         reply["comms"] = comms;
         reply["status"] = "ok";
-        send_reply("comm_info_reply", nl::json::object(), std::move(reply), c);
+        send_reply(request.identities(), "comm_info_reply",request.header(), nl::json::object(), std::move(reply), c);
     }
 
-    void xkernel_core::kernel_info_request(xmessage /* request */, channel c)
+    void xkernel_core::kernel_info_request(xmessage request, channel c)
     {
         nl::json reply = p_interpreter->kernel_info_request();
         reply["protocol_version"] = get_protocol_version();
-        send_reply("kernel_info_reply", nl::json::object(), std::move(reply), c);
+        send_reply(request.identities(), "kernel_info_reply", request.header(), nl::json::object(), std::move(reply), c);
     }
 
     void xkernel_core::shutdown_request(xmessage request, channel c)
@@ -334,15 +369,15 @@ namespace xeus
         p_server->stop();
         nl::json reply;
         reply["restart"] = restart;
-        publish_message("shutdown", nl::json::object(), std::move(reply), buffer_sequence(), channel::CONTROL);
-        send_reply("shutdown_reply", nl::json::object(), std::move(reply), c);
+        publish_message("shutdown", request.header(), nl::json::object(), std::move(reply), buffer_sequence(), channel::CONTROL);
+        send_reply(request.identities(), "shutdown_reply", request.header(), nl::json::object(), std::move(reply), c);
     }
 
-    void xkernel_core::interrupt_request(xmessage, channel c)
+    void xkernel_core::interrupt_request(xmessage request, channel c)
     {
         nl::json reply = nl::json::object();
-        publish_message("interrupt", nl::json::object(), std::move(reply), buffer_sequence(), channel::CONTROL);
-        send_reply("interrupt_reply", nl::json::object(), std::move(reply), c);
+        publish_message("interrupt", request.header(), nl::json::object(), std::move(reply), buffer_sequence(), channel::CONTROL);
+        send_reply(request.identities(), "interrupt_reply", request.header(), nl::json::object(), std::move(reply), c);
     }
 
     void xkernel_core::debug_request(xmessage request, channel c)
@@ -351,43 +386,32 @@ namespace xeus
         {
             nl::json reply = p_debugger->process_request(request.header(), request.content());
             nl::json metadata = get_metadata();
-            send_reply("debug_reply", std::move(metadata), std::move(reply), c);
+            send_reply(request.identities(), "debug_reply", request.header(), std::move(metadata), std::move(reply), c);
         }
     }
 
-    void xkernel_core::publish_status(const std::string& status, channel c)
+    void xkernel_core::publish_status(nl::json parent_header, const std::string& status, channel c)
     {
         nl::json content;
         content["execution_state"] = status;
-        publish_message("status", nl::json::object(), std::move(content), buffer_sequence(), c);
+        publish_message("status", parent_header, nl::json::object(), std::move(content), buffer_sequence(), c);
     }
 
-    void xkernel_core::publish_execute_input(const std::string& code,
+    void xkernel_core::publish_execute_input(nl::json parent_header,
+                                             const std::string& code,
                                              int execution_count)
     {
         nl::json content;
         content["code"] = code;
         content["execution_count"] = execution_count;
         publish_message("execute_input",
+                         std::move(parent_header),
                          nl::json::object(),
                          std::move(content),
                          buffer_sequence(),
                          channel::SHELL);
     }
 
-    void xkernel_core::send_reply(const std::string& reply_type,
-                                  nl::json metadata,
-                                  nl::json reply_content,
-                                  channel c)
-    {
-        send_reply(get_parent_id(c),
-                   reply_type,
-                   get_parent_header(c),
-                   std::move(metadata),
-                   std::move(reply_content),
-                   c);
-    }
-
     void xkernel_core::send_reply(const guid_list& id_list,
                                   const std::string& reply_type,
                                   nl::json parent_header,
@@ -440,25 +464,6 @@ namespace xeus
         return metadata;
     }
 
-    void xkernel_core::set_parent(const guid_list& parent_id,
-                                  const nl::json& parent_header,
-                                  channel c)
-    {
-        auto idx = static_cast<std::size_t>(c);
-        m_parent_id[idx] = parent_id;
-        m_parent_header[idx] = parent_header;
-    }
-
-    const xkernel_core::guid_list& xkernel_core::get_parent_id(channel c) const
-    {
-        return m_parent_id[std::size_t(c)];
-    }
-
-    nl::json xkernel_core::get_parent_header(channel c) const
-    {
-        return m_parent_header[std::size_t(c)];
-    }
-
     void xkernel_core::comm_open(xmessage request, channel)
     {
         m_comm_manager.comm_open(std::move(request));
